package modManager;

import redguard.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.*;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.List;

public class RedguardModManager {
    private static final String TITLE = "Redguard Mod Manager";

    // Standard mod changes file names
    public static final String RTX_CHANGES = "RTX Changes.txt";
    public static final String MENU_CHANGES = "Menu Changes.txt";
    public static final String MAP_CHANGES = "Map Changes.txt";

    // Directories and file created by the program
    private static final String MODS_DIRECTORY = "Mods";
    private static final String DATA_DIRECTORY = "data";
    private static final String SETTINGS_FILE = "Settings.txt";
    private static final String MOD_LIST_FILE = "Mod List.txt";

    // Path fields
    private static Path programPath;
    private static Path modsPath;
    private static Path dataPath;
    private static Path gamePath;

    // GUI fields
    private static JFrame window;
    private static ModTable modTable; // Table listing all of the available mods
    private static JTextField nameField, versionField, authorField;
    private static JTextArea descriptionArea;

    // Storage for various mod changes
    private static RtxDatabase rtxDatabase;
    private static MenuFile menuFile;
    private static MapDatabase mapDatabase;

    /**
     * Starting point for the program. Set the look and feel, initialize various paths, load settings from previous
     * sessions, load the mod list, and create the GUI.
     *
     * @param args Not used
     */
    public static void main(String[] args) {
        // Set system look and feel
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (ClassNotFoundException | InstantiationException | IllegalAccessException
                | UnsupportedLookAndFeelException e) {
            e.printStackTrace();
        }

        // Get program path
        try {
            programPath = Paths.get(RedguardModManager.class.getProtectionDomain().getCodeSource().getLocation().toURI()).getParent();
        } catch (URISyntaxException e) {
            e.printStackTrace();
        }
        modsPath = programPath.resolve(MODS_DIRECTORY);

        // Load settings if the file exists. Otherwise, ask for game directory.
        File settingsFile = programPath.resolve(SETTINGS_FILE).toFile();
        if (settingsFile.exists()) {
            loadSettings(settingsFile);
        } else {
            askGameDirectory();
        }
        dataPath = gamePath.resolve(DATA_DIRECTORY);

        // Load mod list
        modTable = new ModTable();
        loadModList();

        // Load common data files
        preloadFiles();

        // Create GUI
        SwingUtilities.invokeLater(RedguardModManager::createAndShowGUI);
    }

    /**
     * Copy files to the data directory and load some common files for use later.
     */
    private static void preloadFiles() {
        // Backup files if the data directory doesn't exist yet
        if (dataPath.toFile().mkdir()) {
            try {
                Files.copy(gamePath.resolve("ENGLISH.RTX"), dataPath.resolve("ENGLISH.RTX"));
                Files.copy(gamePath.resolve("MENU.INI"), dataPath.resolve("MENU.INI"));
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        try {
            rtxDatabase = new RtxDatabase();
            rtxDatabase.readFile(dataPath.resolve("ENGLISH.RTX").toFile());
            menuFile = new MenuFile();
            menuFile.readFile(dataPath.resolve("MENU.INI").toFile());
            mapDatabase = new MapDatabase(rtxDatabase);
            mapDatabase.readWorldFile(gamePath.resolve("WORLD.INI").toFile());
            mapDatabase.readSoupFile(gamePath.resolve("soup386/SOUP386.DEF").toFile());
            mapDatabase.readItemsFile(gamePath.resolve("ITEM.INI").toFile());
        } catch (IOException e) {
            e.printStackTrace();
        }

        // Finish backing up now that we have the list of map files
        if (dataPath.resolve("maps").toFile().mkdir()) {
            for (MapFile mapFile : mapDatabase.getMapFiles()) {
                try {
                    Files.copy(gamePath.resolve(mapFile.getFullName()), dataPath.resolve(mapFile.getFullName()));
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Create all the GUI components for the main mod manager window.
     */
    private static void createAndShowGUI() {
        // Create a window and make sure to save data before exiting
        window = new JFrame(TITLE);
        window.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        window.setMinimumSize(new Dimension(800, 640));
        window.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent event) {
                exitProgram(0);
            }
        });

        // GUI is complex enough to have separate methods for each section
        createMenuBar();
        addModTable();
        createInfoPanel();
        createBottomPanel();

        // Pack the window, center it, and make it visible
        window.pack();
        window.setLocationRelativeTo(null);
        window.setVisible(true);
    }

    /**
     * Create the menu bar at the top with items for various commands.
     */
    private static void createMenuBar() {
        JMenuBar menuBar = new JMenuBar();
        window.setJMenuBar(menuBar);

        // File menu
        JMenu fileMenu = new JMenu("File");
        fileMenu.setMnemonic('f');
        menuBar.add(fileMenu);
        // File menu items
        ModManagerUtils.createMenuItem(fileMenu, "Add mod...", evt -> addMod());
        ModManagerUtils.createMenuItem(fileMenu, "Delete selected mod...", evt -> deleteSelectedMod());
        fileMenu.addSeparator();
        ModManagerUtils.createMenuItem(fileMenu, "Exit", evt -> exitProgram(0));

        // Edit menu
        JMenu editMenu = new JMenu("Edit");
        editMenu.setMnemonic('e');
        menuBar.add(editMenu);
        // Edit menu items
        ModManagerUtils.createMenuItem(editMenu, "Create mod...", evt -> createMod());
        editMenu.addSeparator();
        ModManagerUtils.createMenuItem(editMenu, "Open Subtitle Editor...", evt -> new SubtitleEditor(rtxDatabase));
        ModManagerUtils.createMenuItem(editMenu, "Open Selected Mod in Subtitle Editor...", evt -> openModInSubtitleEditor());
        editMenu.addSeparator();
        ModManagerUtils.createMenuItem(editMenu, "Open Script Editor...", evt -> new ScriptEditor(dataPath, mapDatabase));
        ModManagerUtils.createMenuItem(editMenu, "Open Selected Mod in Script Editor...", evt -> openModInScriptEditor());
        editMenu.addSeparator();
        ModManagerUtils.createMenuItem(editMenu, "Open Texture Viewer...", evt -> new TextureViewer(gamePath));
    }

    /**
     * Adds the mod table to the window with a scroll pane. Also enables mod selections to show up on the info panel.
     */
    private static void addModTable() {
        // When a row is selected, add the mod's information to the info panel
        modTable.getSelectionModel().addListSelectionListener(evt -> loadInfoPanel());

        // Scroll pane for mod table
        JScrollPane tableScroll = new JScrollPane(modTable);
        tableScroll.setPreferredSize(new Dimension(625, 550));
        window.add(tableScroll, BorderLayout.CENTER);
    }

    /**
     * Create info panel that shows name, version, author, and description of selected mod
     */
    private static void createInfoPanel() {
        SpringLayout layout = new SpringLayout();
        JPanel infoPanel = new JPanel(layout);
        infoPanel.setPreferredSize(new Dimension(250, 0));
        infoPanel.setBackground(new Color(230, 230, 230));
        window.add(infoPanel, BorderLayout.EAST);

        // Mod name
        JLabel nameLabel = new JLabel("Name:");
        nameField = new JTextField();
        addLabeledComponent(layout, infoPanel, nameLabel, nameField);
        layout.putConstraint(SpringLayout.NORTH, nameLabel, 5, SpringLayout.NORTH, infoPanel);

        // Mod version
        JLabel versionLabel = new JLabel("Version:");
        versionField = new JTextField();
        addLabeledComponent(layout, infoPanel, versionLabel, versionField);
        layout.putConstraint(SpringLayout.NORTH, versionLabel, 5, SpringLayout.SOUTH, nameField);

        // Mod author
        JLabel authorLabel = new JLabel("Author:");
        authorField = new JTextField();
        addLabeledComponent(layout, infoPanel, authorLabel, authorField);
        layout.putConstraint(SpringLayout.NORTH, authorLabel, 5, SpringLayout.SOUTH, versionField);

        // Mod description
        JLabel descriptionLabel = new JLabel("Description:");
        descriptionArea = new JTextArea(20, 0);
        descriptionArea.setLineWrap(true);
        descriptionArea.setWrapStyleWord(true);
        descriptionArea.setFont(nameField.getFont());
        JScrollPane descriptionScrollPane = new JScrollPane(descriptionArea);
        addLabeledComponent(layout, infoPanel, descriptionLabel, descriptionScrollPane);
        layout.putConstraint(SpringLayout.NORTH, descriptionLabel, 5, SpringLayout.SOUTH, authorField);

        // Save and cancel buttons
        JButton saveButton = new JButton("Save"); // Save button
        saveButton.addActionListener(evt -> saveInfoPanel());
        infoPanel.add(saveButton);
        JButton cancelButton = new JButton("Cancel"); // Cancel button
        cancelButton.addActionListener(evt -> loadInfoPanel());
        infoPanel.add(cancelButton);
        layout.putConstraint(SpringLayout.NORTH, saveButton, 10, SpringLayout.SOUTH, descriptionScrollPane);
        layout.putConstraint(SpringLayout.EAST, saveButton, -5, SpringLayout.HORIZONTAL_CENTER, infoPanel);
        layout.putConstraint(SpringLayout.NORTH, cancelButton, 10, SpringLayout.SOUTH, descriptionScrollPane);
        layout.putConstraint(SpringLayout.WEST, cancelButton, 5, SpringLayout.HORIZONTAL_CENTER, infoPanel);
    }

    private static void addLabeledComponent(SpringLayout layout, JComponent container, JLabel label, JComponent component) {
        container.add(label);
        container.add(component);
        layout.putConstraint(SpringLayout.WEST, label, 5, SpringLayout.WEST, container);
        layout.putConstraint(SpringLayout.WEST, component, 5, SpringLayout.WEST, container);
        layout.putConstraint(SpringLayout.EAST, component, -5, SpringLayout.EAST, container);
        layout.putConstraint(SpringLayout.NORTH, component, 0, SpringLayout.SOUTH, label);
    }

    /**
     * Create the panel at the bottom of the window, containing buttons like move up and move down.
     */
    private static void createBottomPanel() {
        JPanel bottomPanel = new JPanel();
        bottomPanel.setPreferredSize(new Dimension(0, 50));
        window.add(bottomPanel, BorderLayout.SOUTH);

        // Buttons
        ModManagerUtils.createButton(bottomPanel, "Add Mod", evt -> addMod());
        bottomPanel.add(Box.createHorizontalStrut(10));
        ModManagerUtils.createButton(bottomPanel, "Move Up", evt -> modTable.moveSelectedModUp());
        ModManagerUtils.createButton(bottomPanel, "Move Down", evt -> modTable.moveSelectedModDown());
        bottomPanel.add(Box.createHorizontalStrut(10));
        ModManagerUtils.createButton(bottomPanel, "Apply Changes", evt -> applyChanges());
    }

    /**
     * Ask the user where the Redguard directory is found.
     */
    private static void askGameDirectory() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Open Redguard Directory Containing REDGUARD.EXE");
        chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        while (true) {
            int option = chooser.showOpenDialog(window);
            if (option != JFileChooser.APPROVE_OPTION) {
                System.exit(0);
            }
            if (chooser.getSelectedFile().isDirectory()) {
                gamePath = chooser.getSelectedFile().toPath();
                break;
            }
            ModManagerUtils.showError(window, "Please select a directory.");
        }
    }

    /**
     * Load settings file and set associated variables.
     *
     * @param settingsFile The file to save to
     */
    private static void loadSettings(File settingsFile) {
        try (BufferedReader reader = new BufferedReader(new FileReader(settingsFile))) {
            String line = reader.readLine();
            String[] split = line.split(" = ");
            gamePath = Paths.get(split[1]);
            if (!gamePath.toFile().exists()) {
                askGameDirectory();
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Save settings to a file.
     */
    private static void saveSettings() {
        try (BufferedWriter writer = new BufferedWriter(new FileWriter(programPath.resolve(SETTINGS_FILE).toFile()))) {
            writer.write("gamePath = " + gamePath);
            writer.newLine();
        } catch (IOException e) {
            ModManagerUtils.showError(window, "Could not save settings.");
        }
    }

    /**
     * Create the mods directory if it does not exist, load mods from the mod list, and check for new mods in the mods directory.
     */
    private static void loadModList() {
        // Create mods directory if it does not exist
        File modsDirectory = modsPath.toFile();
        modsDirectory.mkdir();

        // Load mod list if it exists
        File modListFile = programPath.resolve(MOD_LIST_FILE).toFile();
        if (modListFile.exists()) {
            try (BufferedReader reader = new BufferedReader(new FileReader(modListFile))) {
                for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                    // Split by tabs: should only have two parts, the first part being whether the mod is enabled, and the second part its name
                    String[] split = line.split("\t");
                    Path modPath = modsPath.resolve(split[1]);
                    if (modPath.toFile().exists()) {
                        // Mod class handles parsing the about file to get version, author, etc.
                        modTable.addMod(new Mod(modPath, Boolean.parseBoolean(split[0])));
                    }
                }
            } catch (IOException e) {
                ModManagerUtils.showError(window, "An error occurred while loading mods in " + MOD_LIST_FILE + ".");
            }
        }

        // Check for new mods
        File[] files = modsDirectory.listFiles();
        if (files != null) {
            for (File f : files) {
                if (modTable.getModList().stream().noneMatch(mod -> mod.getName().equals(f.getName()))) {
                    try {
                        modTable.addMod(new Mod(f.toPath(), false));
                    } catch (IOException e) {
                        ModManagerUtils.showError(window, "An error occurred while loading " + f.getName() + ".");
                    }
                }
            }
        }
    }

    /**
     * Save mod list to a file for next time the mod manager runs.
     */
    private static void saveModList() {
        try (BufferedWriter writer = new BufferedWriter(new FileWriter(programPath.resolve(MOD_LIST_FILE).toFile()))) {
            for (Mod mod : modTable.getModList()) {
                writer.write(mod.isEnabled() + "\t" + mod.getName());
                writer.newLine();
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Update the info panel with details for the current selected mod.
     */
    private static void loadInfoPanel() {
        Mod selectedMod = modTable.getSelectedMod();
        if (selectedMod == null) {
            nameField.setText("");
            versionField.setText("");
            authorField.setText("");
            descriptionArea.setText("");
        } else {
            nameField.setText(selectedMod.getName());
            versionField.setText(selectedMod.getVersion());
            authorField.setText(selectedMod.getAuthor());
            descriptionArea.setText(selectedMod.getDescription());
            descriptionArea.setCaretPosition(0);
        }
    }

    /**
     * Save the mod details in the info panel to the mod's directory.
     */
    private static void saveInfoPanel() {
        Mod selectedMod = modTable.getSelectedMod();
        if (selectedMod != null) {
            // Fix blank and untrimmed names
            String name = nameField.getText().trim();
            if (name.isEmpty()) {
                name = selectedMod.getName();
            }
            nameField.setText(name);

            // Move the mod directory if necessary and save the new information there (handled by Mod class)
            Path path = modsPath.resolve(name);
            if (name.equals(selectedMod.getName()) || !path.toFile().exists()) {
                try {
                    Files.move(modsPath.resolve(selectedMod.getName()), path, StandardCopyOption.REPLACE_EXISTING);
                    selectedMod.saveInfo(modsPath.resolve(name), versionField.getText(), authorField.getText(), descriptionArea.getText());
                } catch (IOException e) {
                    ModManagerUtils.showError(window, "An error occurred while saving mod info.");
                }
            } else {
                ModManagerUtils.showError(window, "That name is already in use.");
            }
            modTable.updateSelectedRow();
        }
    }

    /**
     * Add a mod from a folder that the user selects.
     */
    private static void addMod() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Open Mod Folder");
        chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        while (true) {
            int option = chooser.showOpenDialog(window);
            File file = chooser.getSelectedFile();
            if (option != JFileChooser.APPROVE_OPTION) {
                break;
            }
            if (file.isDirectory()) {
                try {
                    ModManagerUtils.copyDirectory(file.toPath(), modsPath.resolve(file.getName()));
                    Mod mod = new Mod(file.toPath(), false);
                    modTable.addMod(mod);
                } catch (IOException e) {
                    e.printStackTrace();
                }
                break;
            }
            ModManagerUtils.showError(window, "Please select a directory.");
        }
    }

    /**
     * Confirms if the user wants to delete a selected mod, then deletes its files and the mod table entry.
     */
    private static void deleteSelectedMod() {
        Mod mod = modTable.getSelectedMod();
        if (mod != null) {
            File modFile = getModPath(mod).toFile();
            int shouldDelete = JOptionPane.showConfirmDialog(window, "Are you sure you want to delete "
                    + mod.getName() + "?", "Confirm deletion", JOptionPane.YES_NO_OPTION);
            if (shouldDelete == JOptionPane.YES_OPTION) {
                ModManagerUtils.deleteDirectory(modFile);
                modTable.removeSelectedMod();
            }
        }
    }

    /**
     * Collects all of the changes from enabled mods in their load order, and patches the game files.
     */
    private static void applyChanges() {
        // Check if the user really wants to apply changes
        int shouldApply = JOptionPane.showConfirmDialog(window, "Are you sure you want to apply changes from the enabled mods?",
                "Confirm apply changes", JOptionPane.YES_NO_OPTION);
        if (shouldApply != JOptionPane.YES_OPTION) {
            return;
        }

        RtxDatabase modifiedDatabase = null;
        MenuFile modifiedMenu = null;
        MapChanges mapChanges = new MapChanges();
        try {
            // Go through all enabled mods and collect the sum of their changes, based on load order
            for (Mod mod : modTable.getModList()) {
                if (mod.isEnabled()) {
                    Path path = getModPath(mod);

                    // Get RTX changes
                    File rtxChangesFile = path.resolve(RTX_CHANGES).toFile();
                    if (rtxChangesFile.exists()) {
                        if (modifiedDatabase == null) {
                            modifiedDatabase = new RtxDatabase(rtxDatabase);
                        }
                        modifiedDatabase.applyChanges(rtxChangesFile);
                    }

                    // Get menu changes
                    File menuChangesFile = path.resolve(MENU_CHANGES).toFile();
                    if (menuChangesFile.exists()) {
                        modifiedMenu = new MenuFile(menuFile);
                        modifiedMenu.readChanges(menuChangesFile);
                    }

                    // Get map changes
                    File mapChangesFile = path.resolve(MAP_CHANGES).toFile();
                    if (mapChangesFile.exists()) {
                        mapChanges.readChanges(mapChangesFile);
                    }
                }
            }

            // Replace ENGLISH.RTX, either with a new version or the original copy
            if (modifiedDatabase == null) {
                Files.copy(dataPath.resolve("ENGLISH.RTX"), gamePath.resolve("ENGLISH.RTX"), StandardCopyOption.REPLACE_EXISTING);
            } else {
                modifiedDatabase.writeFile(gamePath.resolve("ENGLISH.RTX").toFile());
            }

            // Replace MENU.INI, again with a new version or the original copy
            if (modifiedMenu == null) {
                Files.copy(dataPath.resolve("MENU.INI"), gamePath.resolve("MENU.INI"), StandardCopyOption.REPLACE_EXISTING);
            } else {
                modifiedMenu.writeFile(gamePath.resolve("MENU.INI").toFile());
            }

            // Replace map files that have changes and copy the rest
            for (MapFile mapFile : mapDatabase.getMapFiles()) {
                if (mapChanges.hasModifiedMap(mapFile.getName())) {
                    if (mapFile.isEmpty()) {
                        mapFile.readMap(dataPath.resolve(mapFile.getFullName()).toFile());
                    }
                    String modifiedScript = mapFile.getModifiedScript(mapChanges);
                    mapFile.writeMap(gamePath.resolve(mapFile.getFullName()).toFile(), modifiedScript);
                } else {
                    Files.copy(dataPath.resolve(mapFile.getFullName()), gamePath.resolve(mapFile.getFullName()), StandardCopyOption.REPLACE_EXISTING);
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
        JOptionPane.showMessageDialog(window, "Changes were applied successfully.", "Redguard Mod Manager",
                JOptionPane.INFORMATION_MESSAGE);
    }

    /**
     * Get the path to a mod's directory.
     *
     * @param mod The mod whose path is returned
     * @return The path to the mod
     */
    public static Path getModPath(Mod mod) {
        return modsPath.resolve(mod.getName());
    }

    /**
     * Create an empty mod with a chosen name so that files can be added to it through the edit menu.
     */
    private static void createMod() {
        while (true) {
            String name = JOptionPane.showInputDialog(window, "Enter mod name:", "Create mod", JOptionPane.QUESTION_MESSAGE);
            // User canceled or exited the dialog
            if (name == null) {
                break;
            }
            // Check if name is blank
            String trimmedName = name.trim();
            if (trimmedName.isEmpty()) {
                ModManagerUtils.showError(window, "Name cannot be blank.");
                continue;
            }

            // Create directory if it exists, or throw an error message
            Path path = modsPath.resolve(trimmedName);
            if (path.toFile().mkdir()) {
                modTable.addMod(new Mod(path));
                break;
            } else {
                ModManagerUtils.showError(window, "That name is already in use.");
            }
        }
    }

    /**
     * If a mod is selected, open the subtitle editor and automatically load the mod's changes file.
     */
    public static void openModInSubtitleEditor() {
        Mod mod = modTable.getSelectedMod();
        if (mod != null) {
            File changesFile = getModPath(mod).resolve(RTX_CHANGES).toFile();
            if (changesFile.exists()) {
                SubtitleEditor editor = new SubtitleEditor(rtxDatabase);
                editor.loadChangesFile(changesFile);
            } else {
                JOptionPane.showMessageDialog(window, "Mod does not have an " + RTX_CHANGES + " file.",
                        "No changes file", JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    /**
     * If a mod is selected, open it in the script editor and automatically load all of the mod's map changes files.
     */
    private static void openModInScriptEditor() {
        Mod mod = modTable.getSelectedMod();
        if (mod != null) {
            File changesFile = getModPath(mod).resolve(MAP_CHANGES).toFile();
            if (changesFile.exists()) {
                ScriptEditor editor = new ScriptEditor(dataPath, mapDatabase);
                editor.loadChangesFile(changesFile);
            } else {
                JOptionPane.showMessageDialog(window, "Mod does not have a " + MAP_CHANGES + " file.",
                        "No changes file", JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    /**
     * Shows a list of mod names to choose from.
     *
     * @return The selected mod, or null if there are no mods or the dialog was closed
     */
    public static Mod showModSelectionDialog(Component parentComponent, String message, String title) {
        List<Mod> modList = modTable.getModList();
        if (modList.isEmpty()) {
            return null;
        }
        Object[] modArray = modList.toArray();
        return (Mod) JOptionPane.showInputDialog(parentComponent, message, title,
                JOptionPane.PLAIN_MESSAGE, null, modArray, modArray[0]);
    }

    /**
     * Saves settings and the mod list before exiting.
     *
     * @param status Status to pass to System.exit
     */
    private static void exitProgram(int status) {
        // Save settings and mod list before exiting
        saveSettings();
        saveModList();
        System.exit(status);
    }
}