package modManager;

import diff.Diff;
import redguard.MapChanges;
import redguard.MapDatabase;
import redguard.MapFile;

import javax.swing.*;
import javax.swing.filechooser.FileNameExtensionFilter;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.*;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

public class ScriptEditor {
    private static final String TITLE = "Redguard Script Editor";

    // GUI fields
    private JFrame window;
    private JMenuItem saveMapItem, saveScriptItem, saveChangesItem;
    private JTabbedPane tabbedPane;
    private List<MapFile> mapFileTabs;
    private MapFile currentMapFile;
    private JTextField findTextField;

    // Other fields
    private Path dataPath;
    private MapDatabase mapDatabase;

    public ScriptEditor(Path dataPath, MapDatabase mapDatabase) {
        this.dataPath = dataPath;
        this.mapDatabase = mapDatabase;
        createAndShowGUI();
    }

    private void createAndShowGUI() {
        // Remove focus border around tabs
        UIManager.put("TabbedPane.focus", new Color(0, 0, 0, 0));

        window = new JFrame(TITLE);
        window.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        window.setMinimumSize(new Dimension(800, 600));

        // Split GUI creation into sections
        createMenuBar();
        createTabbedPane();
        createBottomPanel();

        window.pack();
        window.setLocationRelativeTo(null);
        window.setVisible(true);
    }

    /**
     * Create the menu bar at the top with items for various commands.
     */
    private void createMenuBar() {
        JMenuBar menuBar = new JMenuBar();
        window.setJMenuBar(menuBar);

        // File menu
        JMenu fileMenu = new JMenu("File");
        fileMenu.setMnemonic('f');
        menuBar.add(fileMenu);
        // File menu items
        saveMapItem = ModManagerUtils.createMenuItem(fileMenu, "Save Map...", evt -> saveMapFile());
        saveMapItem.setEnabled(false);
        fileMenu.addSeparator();
        ModManagerUtils.createMenuItem(fileMenu, "Load Script...", evt -> loadScriptFile());
        saveScriptItem = ModManagerUtils.createMenuItem(fileMenu, "Save Script...", evt -> saveScriptFile());
        saveScriptItem.setEnabled(false);
        fileMenu.addSeparator();
        ModManagerUtils.createMenuItem(fileMenu, "Load Changes...", evt -> loadChangesFile());
        saveChangesItem = ModManagerUtils.createMenuItem(fileMenu, "Save Changes...", evt -> saveChangesFile());
        saveChangesItem.setEnabled(false);
        fileMenu.addSeparator();
        ModManagerUtils.createMenuItem(fileMenu, "Exit Script Editor", evt -> window.dispose());
    }

    /**
     * Set up the tabbed pane containing tabs and text areas.
     */
    private void createTabbedPane() {
        tabbedPane = new JTabbedPane();
        tabbedPane.setPreferredSize(new Dimension(800, 560));
        tabbedPane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
        tabbedPane.addChangeListener(evt -> {
            int index = tabbedPane.getSelectedIndex();
            currentMapFile = (index == -1) ? null : mapFileTabs.get(index);

            boolean canSave = currentMapFile != null;
            saveMapItem.setEnabled(canSave);
            saveScriptItem.setEnabled(canSave);
            saveChangesItem.setEnabled(canSave);
        });
        window.add(tabbedPane, BorderLayout.CENTER);
        mapFileTabs = new ArrayList<>();
    }

    /**
     * Create bottom panel that holds the find text field and load/save buttons
     */
    private void createBottomPanel() {
        JPanel bottomPanel = new JPanel();
        bottomPanel.setPreferredSize(new Dimension(0, 40));
        window.add(bottomPanel, BorderLayout.SOUTH);

        // Find text field
        JLabel findLabel = new JLabel("Find");
        bottomPanel.add(findLabel);
        findTextField = new JTextField(10);
        findTextField.addActionListener(evt -> findText());
        bottomPanel.add(findTextField);

        // Buttons
        ModManagerUtils.createButton(bottomPanel, "Load Map", evt -> loadMapFile());
        ModManagerUtils.createButton(bottomPanel, "Load from Mod", evt -> loadFromMod());
        ModManagerUtils.createButton(bottomPanel, "Save to Mod", evt -> saveToMod());
    }

    /**
     * Try to find the text in findTextField within the current script's text.
     */
    private void findText() {
        if (currentMapFile != null) {
            String find = findTextField.getText().toLowerCase();
            if (!find.isEmpty()) {
                JTextArea textArea = getSelectedTextArea();
                String text = textArea.getText().toLowerCase();
                int pos = textArea.getCaretPosition();
                int index = text.substring(pos).indexOf(find);
                if (index == -1) {
                    index = text.substring(0, pos).indexOf(find);
                    pos = 0;
                }
                if (index >= 0) {
                    textArea.requestFocus();
                    textArea.setCaretPosition(pos + index + find.length());
                    textArea.select(pos + index, pos + index + find.length());
                    findTextField.requestFocus();
                    JScrollBar vBar = getSelectedScrollPane().getVerticalScrollBar();
                    vBar.setValue(vBar.getValue() + textArea.getHeight());
                }
            }
        }
    }

    private JScrollPane getScrollPaneAt(int index) {
        return (JScrollPane) tabbedPane.getComponentAt(index);
    }

    private JScrollPane getSelectedScrollPane() {
        return (JScrollPane) tabbedPane.getSelectedComponent();
    }

    private JTextArea getTextAreaAt(int index) {
        return (JTextArea) getScrollPaneAt(index).getViewport().getView();
    }

    private JTextArea getSelectedTextArea() {
        return (JTextArea) getSelectedScrollPane().getViewport().getView();
    }

    /**
     * Add a new tab or replace an existing tab with the same name, then put the given script into the new tab's text
     * area. Associate this tab with the given map file in mapFileTabs.
     *
     * @param mapFile The map file which names this tab and which determines how to handle the script
     * @param script  The script to add to the tab's text area
     */
    private void addTab(MapFile mapFile, String script) {
        currentMapFile = mapFile;

        JTextArea textArea;
        int tabIndex = mapFileTabs.indexOf(currentMapFile);
        if (tabIndex == -1) {
            // Create new tab
            tabIndex = mapFileTabs.size();
            mapFileTabs.add(currentMapFile);

            textArea = new JTextArea();
            textArea.setLineWrap(true);
            textArea.setWrapStyleWord(true);
            textArea.setMargin(new Insets(5, 5, 5, 5));

            tabbedPane.addTab(currentMapFile.getName(), new JScrollPane(textArea));
            createTabButton(tabIndex);
        } else {
            // Tab exists
            textArea = getTextAreaAt(tabIndex);
        }
        textArea.setText(script);
        textArea.setCaretPosition(0); // Move back to the top
        tabbedPane.setSelectedIndex(tabIndex);
    }

    private void createTabButton(int index) {
        Box tabBox = Box.createHorizontalBox();
        tabBox.add(new JLabel(currentMapFile.getName()));
        tabBox.add(Box.createHorizontalStrut(10));

        JButton closeButton = new JButton(" X ");
        closeButton.setBorderPainted(false);
        closeButton.setMargin(new Insets(0, 0, 0, 0));
        closeButton.setFocusable(false);
        closeButton.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseEntered(MouseEvent e) {
                closeButton.setBorderPainted(true);
            }

            @Override
            public void mouseExited(MouseEvent e) {
                closeButton.setBorderPainted(false);
            }
        });
        closeButton.addActionListener(evt -> removeTab(tabbedPane.indexOfTabComponent(tabBox)));
        tabBox.add(closeButton);
        tabbedPane.setTabComponentAt(index, tabBox);
    }

    private void removeTab(int index) {
        tabbedPane.removeTabAt(index);
        mapFileTabs.remove(index);
    }

    private void loadMapFile() {
        Object[] mapFiles = mapDatabase.getMapFiles().toArray();
        MapFile selection = (MapFile) JOptionPane.showInputDialog(window, "Choose a map to load.",
                "Load Map", JOptionPane.PLAIN_MESSAGE, null, mapFiles, mapFiles[0]);
        if (selection != null) {
            try {
                selection.readMap(dataPath.resolve(selection.getFullName()).toFile());
            } catch (IOException e) {
                e.printStackTrace();
            }
            addTab(selection, selection.getScript());
        }
    }

    private void saveMapFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Save Map");
        chooser.setFileFilter(new FileNameExtensionFilter("Map files (*.RGM)", "RGM"));
        chooser.setAcceptAllFileFilterUsed(false);
        chooser.setSelectedFile(new File(currentMapFile.getName()));
        int option = chooser.showSaveDialog(window);
        if (option == JFileChooser.APPROVE_OPTION) {
            File file = ModManagerUtils.forceFileExtension(chooser.getSelectedFile(), ".RGM");
            if (ModManagerUtils.confirmReplace(window, file)) {
                try {
                    currentMapFile.writeMap(file, getSelectedTextArea().getText());
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    private void loadScriptFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Load Script");
        chooser.setFileFilter(new FileNameExtensionFilter("Normal text files (*.txt)", "txt"));
        chooser.setAcceptAllFileFilterUsed(false);
        int option = chooser.showOpenDialog(window);
        if (option == JFileChooser.APPROVE_OPTION && chooser.getSelectedFile().getName().endsWith(".txt")) {
            try (BufferedReader reader = new BufferedReader(new FileReader(chooser.getSelectedFile()))) {
                StringBuilder script = new StringBuilder();
                String name = reader.readLine();
                script.append(name);
                for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                    script.append("\n").append(line);
                }
                MapFile mapFile = mapDatabase.getMapFileFromName(name);
                mapFile.readMap(dataPath.resolve(mapFile.getFullName()).toFile());
                addTab(mapFile, script.toString());
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    private void saveScriptFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Save Script");
        chooser.setFileFilter(new FileNameExtensionFilter("Normal text files (*.txt)", "txt"));
        chooser.setAcceptAllFileFilterUsed(false);
        chooser.setSelectedFile(new File(currentMapFile.getName()));
        int option = chooser.showSaveDialog(window);
        if (option == JFileChooser.APPROVE_OPTION) {
            File file = ModManagerUtils.forceFileExtension(chooser.getSelectedFile(), ".txt");
            if (ModManagerUtils.confirmReplace(window, file)) {
                try (BufferedWriter writer = new BufferedWriter(new FileWriter(file))) {
                    writer.write(getSelectedTextArea().getText());
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    private void loadChangesFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Load Changes");
        chooser.setFileFilter(new FileNameExtensionFilter("Normal text files (*.txt)", "txt"));
        chooser.setAcceptAllFileFilterUsed(false);
        int option = chooser.showOpenDialog(window);
        if (option == JFileChooser.APPROVE_OPTION) {
            loadChangesFile(chooser.getSelectedFile());
        }
    }

    public void loadChangesFile(File fileToLoad) {
        MapChanges mapChanges = new MapChanges();
        try {
            mapChanges.readChanges(fileToLoad);
        } catch (IOException e) {
            e.printStackTrace();
        }

        for (MapFile mapFile : mapDatabase.getMapFiles()) {
            if (mapChanges.hasModifiedMap(mapFile.getName())) {
                try {
                    mapFile.readMap(dataPath.resolve(mapFile.getFullName()).toFile());
                } catch (IOException e) {
                    e.printStackTrace();
                }
                String modifiedScript = mapFile.getModifiedScript(mapChanges);
                addTab(mapFile, modifiedScript);
            }
        }
    }

    private void saveChangesFile() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Save Changes");
        chooser.setFileFilter(new FileNameExtensionFilter("Normal text files (*.txt)", "txt"));
        chooser.setAcceptAllFileFilterUsed(false);
        chooser.setSelectedFile(new File("Map Changes"));
        int option = chooser.showSaveDialog(window);
        if (option == JFileChooser.APPROVE_OPTION) {
            File file = ModManagerUtils.forceFileExtension(chooser.getSelectedFile(), ".txt");
            if (ModManagerUtils.confirmReplace(window, file)) {
                saveChangesFile(file);
            }
        }
    }

    private void saveChangesFile(File fileToSave) {
        MapChanges mapChanges = new MapChanges();
        for (int i = 0; i < mapFileTabs.size(); i++) {
            MapFile mapFile = mapFileTabs.get(i);
            String[] originalLines = mapFile.getScript().split("\n");
            String[] modifiedLines = getTextAreaAt(i).getText().split("\n");
            mapChanges.addChanges(mapFile.getName(), Diff.diff(originalLines, modifiedLines));
        }
        try {
            mapChanges.writeChanges(fileToSave);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void loadFromMod() {
        Mod mod = RedguardModManager.showModSelectionDialog(window, "Choose a mod to load script changes.",
                "Load from Mod");
        if (mod != null) {
            File changesFile = RedguardModManager.getModPath(mod).resolve(RedguardModManager.MAP_CHANGES).toFile();
            if (changesFile.exists()) {
                loadChangesFile(changesFile);
            } else {
                JOptionPane.showMessageDialog(window, "Mod does not have a " + RedguardModManager.MAP_CHANGES
                        + " file.", "No changes file", JOptionPane.INFORMATION_MESSAGE);
            }
        }
    }

    private void saveToMod() {
        Mod mod = RedguardModManager.showModSelectionDialog(window, "Choose a mod to save script changes.",
                "Save to Mod");
        if (mod != null) {
            File changesFile = RedguardModManager.getModPath(mod).resolve(RedguardModManager.MAP_CHANGES).toFile();
            saveChangesFile(changesFile);
            if (changesFile.length() == 0) {
                changesFile.delete();
            }
        }
    }
}