package modManager;

import redguard.Texture;
import redguard.TextureFile;

import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.Timer;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.util.*;
import java.util.List;

public class TextureViewer {
    private static final String TITLE = "Redguard Texture Viewer";
    private static String colorMap = "REDGUARD.COL";
    private static Set<String> ignoredGXA = new HashSet<>();

    static {
        ignoredGXA.add("GXICONS.GXA");
        ignoredGXA.add("INVBACK.GXA");
        ignoredGXA.add("INVMASK.GXA");
        ignoredGXA.add("PICKBLOB.GXA");
        ignoredGXA.add("SNUFF.GXA");
    }

    // GUI fields
    private JFrame window;
    private JList<String> fileList;

    // Other fields
    private Path gamePath;
    private TextureFile textureFile;
    private int textureIndex;
    private int frameIndex;

    public TextureViewer(Path gamePath) {
        this.gamePath = gamePath;
        try {
            TextureFile.loadDefaultCMap(gamePath.resolve("fxart/" + colorMap).toFile());
        } catch (IOException e) {
            e.printStackTrace();
        }
        createAndShowGUI();
    }

    private void createAndShowGUI() {
        window = new JFrame(TITLE);
        window.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        window.setMinimumSize(new Dimension(800, 600));

        JPanel rightPanel = new JPanel(new BorderLayout());
        window.add(rightPanel, BorderLayout.CENTER);

        TexturePanel texturePanel = new TexturePanel();
        rightPanel.add(texturePanel, BorderLayout.CENTER);

        JPanel bottomPanel = new JPanel();
        bottomPanel.setPreferredSize(new Dimension(0, 40));
        rightPanel.add(bottomPanel, BorderLayout.SOUTH);

        List<String> comboBoxData = new ArrayList<>();
        File[] fxartFiles = gamePath.resolve("fxart").toFile().listFiles();
        if (fxartFiles != null) {
            for (File f : fxartFiles) {
                if (f.getName().toUpperCase().endsWith(".COL") && f.length() > 0) {
                    comboBoxData.add(f.getName());
                }
            }
        }
        JComboBox<String> colorMapBox = new JComboBox<>(comboBoxData.toArray(new String[0]));
        colorMapBox.setSelectedItem("REDGUARD.COL");
        colorMapBox.addItemListener(evt -> {
            try {
                TextureFile.loadDefaultCMap(gamePath.resolve("fxart/" + colorMapBox.getSelectedItem()).toFile());
            } catch (IOException e) {
                e.printStackTrace();
            }
            if (textureFile.getType() == TextureFile.Type.TEXBSI) {
                loadTextureFile(fileList.getSelectedValue());
            }
        });
        bottomPanel.add(colorMapBox);

        bottomPanel.add(Box.createHorizontalStrut(20));
        ModManagerUtils.createButton(bottomPanel, "<", evt -> setTextureIndex(textureIndex - 1));
        ModManagerUtils.createButton(bottomPanel, ">", evt -> setTextureIndex(textureIndex + 1));
        bottomPanel.add(Box.createHorizontalStrut(20));
        ModManagerUtils.createButton(bottomPanel, "Import Image", evt -> importImage());
        ModManagerUtils.createButton(bottomPanel, "Export Image", evt -> exportImage());
        ModManagerUtils.createButton(bottomPanel, "Save Texture", evt -> writeTexture());
        bottomPanel.add(Box.createHorizontalStrut(20));

        createListPanel();

        window.pack();
        window.setLocationRelativeTo(null);
        window.setVisible(true);
    }

    private void createListPanel() {
        List<String> listData = new ArrayList<>();

        File[] fxartFiles = gamePath.resolve("fxart").toFile().listFiles();
        if (fxartFiles != null) {
            for (File f : fxartFiles) {
                if (f.getName().toUpperCase().startsWith("TEXBSI")) {
                    listData.add(f.getName());
                }
            }
        }
        File[] systemFiles = gamePath.resolve("system").toFile().listFiles();
        if (systemFiles != null) {
            for (File f : systemFiles) {
                String upperName = f.getName().toUpperCase();
                if (upperName.endsWith(".GXA") && !ignoredGXA.contains(upperName)) {
                    listData.add(f.getName());
                }
            }
        }
        File[] fontsFiles = gamePath.resolve("fonts").toFile().listFiles();
        if (fontsFiles != null) {
            for (File f : fontsFiles) {
                if (f.getName().toUpperCase().endsWith(".FNT")) {
                    listData.add(f.getName());
                }
            }
        }

        fileList = new JList<>(listData.toArray(new String[0]));
        fileList.setSelectedIndex(0);
        loadTextureFile(listData.get(0));
        fileList.addListSelectionListener(evt -> loadTextureFile(fileList.getSelectedValue()));

        JScrollPane listScrollPane = new JScrollPane(fileList);
        listScrollPane.setPreferredSize(new Dimension(100, 0));
        window.add(listScrollPane, BorderLayout.WEST);
    }

    private void loadTextureFile(String filename) {
        String name = filename.toUpperCase();
        TextureFile.Type fileType;
        String folder;
        if (name.endsWith(".GXA")) {
            fileType = TextureFile.Type.GXA;
            folder = "system";
        } else if (name.endsWith(".FNT")) {
            fileType = TextureFile.Type.FNT;
            folder = "fonts";
        } else {
            fileType = TextureFile.Type.TEXBSI;
            folder = "fxart";
        }

        textureFile = new TextureFile(fileType);
        try {
            textureFile.loadTextures(gamePath.resolve(folder).resolve(filename).toFile());
        } catch (IOException e) {
            e.printStackTrace();
        }

        setTextureIndex(0);
        frameIndex = 0;
    }

    private void setTextureIndex(int value) {
        textureIndex = value;

        List<Texture> textures = textureFile.getTextures();
        int max = textures.size() - 1;
        if (textureIndex < 0) {
            textureIndex = max;
        } else if (textureIndex > max) {
            textureIndex = 0;
        }

        frameIndex = 0;
    }

    private void importImage() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Import Image from PNG or Folder");
        chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        int option = chooser.showOpenDialog(window);
        if (option == JOptionPane.YES_OPTION) {
            Color[] cMap = textureFile.getTextures().get(textureIndex).getCMap();
            byte[] headerBytes = textureFile.getTextures().get(textureIndex).getHeaderBytes();
            try {
                if (chooser.getSelectedFile().isDirectory()) {
                    File[] files = chooser.getSelectedFile().listFiles();
                    List<BufferedImage> images = new ArrayList<>();
                    if (files != null) {
                        Arrays.sort(files);
                        for (File file : files) {
                            if (file.getName().endsWith(".png")) {
                                images.add(ImageIO.read(file));
                            }
                        }
                    }
                    BufferedImage[] imageArray = images.toArray(new BufferedImage[images.size()]);
                    textureFile.getTextures().set(textureIndex, new Texture(headerBytes, cMap, imageArray));
                } else {
                    BufferedImage image = ImageIO.read(chooser.getSelectedFile());
                    textureFile.getTextures().set(textureIndex, new Texture(headerBytes, cMap, image));
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    private void exportImage() {
        Texture texture = textureFile.getTextures().get(textureIndex);
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Export Image as PNG");

        String[] split = fileList.getSelectedValue().split("\\.");
        if (split[1].equals("GXA") || split[1].equals("FNT")) {
            chooser.setSelectedFile(new File(split[0]));
        } else {
            chooser.setSelectedFile(new File(split[0] + " " + split[1]));
        }

        int option = chooser.showSaveDialog(window);
        if (option == JOptionPane.YES_OPTION) {
            try {
                if (texture.getNumFrames() == 1) {
                    File file = ModManagerUtils.forceFileExtension(chooser.getSelectedFile(), ".png");
                    if (ModManagerUtils.confirmReplace(window, file)) {
                        ImageIO.write(texture.getImages()[0], "png", file);
                    }
                } else {
                    File file = ModManagerUtils.forceFileExtension(chooser.getSelectedFile(), "");
                    if (ModManagerUtils.confirmReplace(window, file)) {
                        file.mkdir();
                        for (int frame = 0; frame < texture.getNumFrames(); frame++) {
                            String str = String.format("%s/Frame %02d.png", file.getName(), frame + 1);
                            ImageIO.write(texture.getImages()[frame], "png", new File(str));
                        }
                    }
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    private void writeTexture() {
        JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Save Texture");
        chooser.setSelectedFile(new File(fileList.getSelectedValue()));
        int option = chooser.showSaveDialog(window);
        if (option == JOptionPane.YES_OPTION) {
            try {
                textureFile.writeTextures(chooser.getSelectedFile());
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    private class TexturePanel extends JPanel {
        public TexturePanel() {
            setPreferredSize(new Dimension(700, 560));
            setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 10));
            new Timer(250, evt -> repaint()).start();
        }

        @Override
        public void paintComponent(Graphics g) {
            super.paintComponent(g);

            if (textureFile != null) {
                Texture texture = textureFile.getTextures().get(textureIndex);
                BufferedImage image = texture.getImages()[frameIndex];

                frameIndex++;
                if (frameIndex >= texture.getNumFrames()) {
                    frameIndex = 0;
                }

                int width, height;
                if (image.getWidth() > image.getHeight()) {
                    width = 500;
                    height = width * image.getHeight() / image.getWidth();
                } else {
                    height = 400;
                    width = height * image.getWidth() / image.getHeight();
                }
                int x = (getWidth() - width) / 2;
                int y = (getHeight() - height) / 2;
                g.drawImage(image, x, y, width, height, null);

                // Details in top-left corner
                g.drawString("Image: " + (textureIndex + 1) + " / " + textureFile.getTextures().size(), 5, 15);
                g.drawString("Size: " + texture.getWidth() + " x " + texture.getHeight() + " px", 5, 30);
                g.drawString("Frames: " + texture.getNumFrames(), 5, 45);
            }
        }
    }
}