package redguard;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

public class RtxDatabase {
    private static final byte[] AUDIO_NO = {0, 0};
    private static final byte[] AUDIO_YES = {0, 1};

    private List<RtxEntry> rtxEntries;

    public RtxDatabase() {
        rtxEntries = new ArrayList<>();
    }

    public RtxDatabase(RtxDatabase other) {
        rtxEntries = new ArrayList<>(other.rtxEntries);
    }

    public List<RtxEntry> getRtxEntries() {
        return rtxEntries;
    }

    public void readFile(File fileToRead) throws IOException {
        rtxEntries.clear();
        DataInputStream input = new DataInputStream(new FileInputStream(fileToRead));
        while (true) {
            String label = Utils.readString(input, 4);
            if (label.equals("END ")) {
                break;
            }

            int length = input.readInt();
            boolean hasAudio = input.readShort() == 1;

            int subtitleLength = Utils.readLittleEndianInt(input);
            String subtitle = Utils.readString(input, subtitleLength);

            byte[] audioBytes = null;
            if (hasAudio) {
                audioBytes = Utils.readBytes(input, length - 6 - subtitleLength);
            }
            rtxEntries.add(new RtxEntry(label, subtitle, audioBytes));
        }
        input.close();
    }

    public void writeFile(File fileToWrite) throws IOException {
        DataOutputStream output = new DataOutputStream(new FileOutputStream(fileToWrite));
        int totalBytes = 0;
        int[] entryPositions = new int[rtxEntries.size()];
        for (int i = 0; i < rtxEntries.size(); i++) {
            RtxEntry entry = rtxEntries.get(i);
            output.writeBytes(entry.getLabel());
            output.writeInt(entry.getLength());
            totalBytes += 8;
            entryPositions[i] = totalBytes;
            output.write(entry.hasAudio() ? AUDIO_YES : AUDIO_NO);
            Utils.writeLittleEndianInt(output, entry.getSubtitleLength());
            output.writeBytes(entry.getSubtitle());
            totalBytes += 6 + entry.getSubtitleLength();
            if (entry.hasAudio()) {
                output.write(entry.getAudioBytes());
                totalBytes += entry.getAudioBytes().length;
            }
        }
        output.writeBytes("END ");
        totalBytes += 4;
        for (int i = rtxEntries.size() - 1; i >= 0; i--) {
            RtxEntry entry = rtxEntries.get(i);
            output.writeBytes(entry.getLabel());
            Utils.writeLittleEndianInt(output, entryPositions[i]);
            Utils.writeLittleEndianInt(output, entry.getLength());
        }
        output.writeBytes("RNAV");
        Utils.writeLittleEndianInt(output, totalBytes);
        Utils.writeLittleEndianInt(output, rtxEntries.size());
        output.close();
    }

    public void applyChanges(File changesFile) throws IOException {
        BufferedReader reader = new BufferedReader(new FileReader(changesFile));
        for (String line = reader.readLine(); line != null; line = reader.readLine()) {
            String[] entry = line.split("\t");
            try {
                int index = Integer.parseInt(entry[0]);
                rtxEntries.get(index).setSubtitle(entry[1]);
            } catch (NumberFormatException e) {
                rtxEntries.add(new RtxEntry(entry[0], entry[1]));
            }
        }
        reader.close();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (RtxEntry entry : rtxEntries) {
            sb.append(entry.getLabel()).append(": ").append(entry.getSubtitle()).append("\n");
        }
        return sb.toString();
    }
}